-------------------------------------------------------------------------------
-- Floating-point sine/cosine (dual operator)
--
-- Generics:
--   - wE [positive] : exponent bit-width
--   - wF [positive] : fraction bit-width
--
-- Ports:
--   - fpX [in]  : operand X in FP format
--   - fpS [out] : sine of X in FP format
--   - fpC [out] : cosine of X in FP format
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
library fplib;
use fplib.pkg_common.all;
use fplib.pkg_fp_common.all;
use fplib.pkg_fp_sin_cos.all;

entity fp_sin_cos is
  generic ( wE : positive := 6;
            wF : positive := 13 );
  port ( fpX : in  std_logic_vector(wE+wF+2 downto 0);
         fpS : out std_logic_vector(wE+wF+2 downto 0);
         fpC : out std_logic_vector(wE+wF+2 downto 0) );
end entity;

architecture arch of fp_sin_cos is
  signal eX : std_logic_vector(wE downto 0);
  signal xX : std_logic_vector(1 downto 0);

  constant gK  : positive := fp_sin_cos_4oPi_g(wE, wF);
  constant nK0 : std_logic_vector(3+2*wF+2**(wE-1)-1+gK+6 downto 0) :=
                                                       fp_sin_cos_4oPi(wE, wF);
  signal   nK1 : std_logic_vector(3+2*wF+2**wE-2+gK+5 downto 0);
  signal   nK2 : std_logic_vector(3+2*wF+2**(wE-1)-1+gK+5 downto 0);
  signal   nK  : std_logic_vector(3+2*wF+gK+5 downto 0);

  signal nY0 : std_logic_vector(3*wF+gK+9 downto 0);
  signal nY  : std_logic_vector(2+wF+gK+5 downto 0);

  signal far_nZ0 : std_logic_vector(wF+gK+4 downto 0);
  signal far_nZ  : std_logic_vector(wF+1 downto 0);
  signal far_eZ0 : std_logic_vector(log2(gK) downto 0);
  signal far_eZ  : std_logic_vector(wE-1 downto 0);
  signal far_mZ0 : std_logic_vector(wF+gK+4 downto 0);
  signal far_mZ  : std_logic_vector(wF+4 downto 0);

  signal cls_nZ0 : std_logic_vector(wF+1+2**(wE-1)-1 downto 0);
  signal cls_nZ  : std_logic_vector(wF+1 downto 0);
  signal cls_eZ0 : std_logic_vector(wE-1 downto 0);
  signal cls_eZ  : std_logic_vector(wE-1 downto 0);
  signal cls_mZ  : std_logic_vector(wF+4 downto 0);

  signal nZ : std_logic_vector(wF+1 downto 0);
  signal eZ : std_logic_vector(wE-1 downto 0);
  signal mZ : std_logic_vector(wF+4 downto 0);

  constant nPio4 : std_logic_vector(wF+4 downto 0) := fp_sin_cos_Pio4(wF);
  
  signal nS0 : std_logic_vector(wF+2 downto 0);
  signal nS1 : std_logic_vector(wF+4 downto 0);
  signal mS2 : std_logic_vector(2*wF+9 downto 0);
  signal mS3 : std_logic_vector(wF+3 downto 0);
  signal mS4 : std_logic_vector(wF+2 downto 0);
  signal eS4 : std_logic_vector(wE downto 0);
  signal mS5 : std_logic_vector(wF downto 0);
  signal eS5 : std_logic_vector(wE downto 0);
  signal eS  : std_logic_vector(wE-1 downto 0);
  signal fS  : std_logic_vector(wF-1 downto 0);

  signal nC0 : std_logic_vector(wF+2 downto 0);
  signal mC1 : std_logic_vector(wF+3 downto 0);
  signal mC2 : std_logic_vector(wF+2 downto 0);
  signal mC3 : std_logic_vector(wF downto 0);
  signal eC  : std_logic_vector(wE-1 downto 0);
  signal fC  : std_logic_vector(wF-1 downto 0);

begin --------------------------------------------------------- Range reduction

  eX <= ("0" & fpX(wE+wF-1 downto wF)) - ("00" & (wE-2 downto 1 => '1') & "0");

  shift_nk0 : left_shift
    generic map ( w => nK0'high,
                  n => wE-1 )
    port map ( i => nK0(nK0'high downto 1),
               s => eX(wE-2 downto 0),
               o => nK1 );

  nK2 <= nK1(nK1'high downto 2**(wE-1)-1);

  nK <= nK2(3+2*wF+gK+5 downto 0)                   when eX(wE-1) = '0' else
        nK0(3+2*wF+gK+5 downto 0)                   when eX(wE) = '0'   else
        nK0(nK0'high downto nK0'high-(3+2*wF+gK+5));

  nY0 <= ("1" & fpX(wF-1 downto 0)) * nK;
  nY  <= nY0(3+2*wF+gK+5 downto wF+1);

  --------------------------------------------------- Range reduction, far path
  
  far_nZ0 <= nY(wF+gK+4 downto 0)                   when nY(wF+gK+5) = '0' else
             (wF+gK+4 downto 0 => '0') - nY(wF+gK+4 downto 0);
  far_nZ <= far_nZ0(wF+gK+4 downto gK+3);

  lzc_norm_far_nz : lzc_norm
    generic map ( w => wF+gK+5,
                  n => log2(gK)+1 )     -- Why +1?
    port map ( i => far_nZ0,
               z => far_eZ0,
               o => far_mZ0 );

  far_eZ <= ((wE-1 downto log2(gK)+1 => '0') & far_eZ0) +
            ((wE-1 downto 1 => '0') & "1");

  far_mZ <= far_mZ0(wF+gK+4 downto gK);

  ------------------------------------------------- Range reduction, close path

  cls_mZ <= nY(wF+gK+4 downto gK)   when nY(wF+gK+5) = '0' else
            nY(wF+gK+5 downto gK+1);

  cls_eZ0 <= (wE-1 downto 0 => '0') - eX(wE-1 downto 0);
  cls_eZ  <= cls_eZ0 + ((wE-1 downto 1 => '0') & (not nY(wF+gK+5)));
  
  shift_cls_nz : right_shift
    generic map ( w => wF+2,
                  n => wE-1 )
    port map ( i => nY(wF+gK+5 downto gK+4),
               s => cls_eZ0(wE-2 downto 0),
               o => cls_nZ0 );

  cls_nZ <= cls_nZ0(wF+2**(wE-1)-1 downto -1+2**(wE-1)-1);

  ----------------------------------------------------- Range reduction, select

  nZ <= far_nZ when eX(wE-1) = '0' else
        cls_nZ;

  eZ <= far_eZ when eX(wE-1) = '0' else
        cls_eZ;

  mZ <= far_mZ when eX(wE-1) = '0' else
        cls_mZ;

  ------------------------------------------------------------------------ Sine

  fsin_nz : fp_sin_cos_fsin
    generic map ( wF => wF )
    port map ( x => nZ,
               r => nS0 );

  nS1 <= nPio4 - ("00" & nS0);

  mS2 <= nS1 * mZ;
  
  mS3(wF+3 downto 1) <= mS2(2*wF+9 downto wF+7);
  mS3(0) <= '0' when mS2(wF+6 downto 0) = (wF+6 downto 0 => '0') else
            '1';

  mS4 <= mS3(wF+3 downto 2) & (mS3(1) or mS3(0)) when mS3(wF+3) = '1' else
         mS3(wF+2 downto 0);

  eS4 <= ("00" & (wE-2 downto 1 => '1') & mS3(wF+3)) - ("0" & eZ);

  round_sin : fp_round
    generic map ( wE => wE,
                  wF => wF )
    port map ( eX => eS4,
               mX => mS4,
               eR => eS5,
               mR => mS5 );

  eS <= eS5(wE-1 downto 0);
  fS <= mS5(wF-1 downto 0);

  ---------------------------------------------------------------------- Cosine

  fcos_nz : fp_sin_cos_fcos
    generic map ( wF => wF )
    port map ( x => nZ,
               r => nC0 );

  mC1 <= ("1" & (wF+2 downto 0 => '0')) - ("0" & nC0);

  mC2 <= mC1(wF+3 downto 2) & (mC1(1) or mC1(0)) when mC1(wF+3) = '1' else
         mC1(wF+2 downto 0);

  mC3 <= mC2(wF+2 downto 2) + ((wF downto 1 => '0') & (mC2(1) and (mC2(2) or mC2(0))));

  eC <= "0" & (wE-2 downto 1 => '1') & (mC1(wF+3) or not mC3(wF));
  fC <= mC3(wF-1 downto 0);

  -----------------------------------------------------------------------------

  xX <= fpX(wE+wF+2 downto wE+wF+1);

  fpS(wE+wF+2 downto wE+wF+1) <= "00"       when xX = "00" else
                                 xX(1) & "1";

  fpS(wE+wF) <= fpX(wE+wF) xor nY(2+wF+gK+5);

  fpS(wE+wF-1 downto 0) <= eS & fS when nY(wF+gK+5) = nY(1+wF+gK+5) or eX(wE) = '1' else
                           eC & fC;

  fpC(wE+wF+2 downto wE+wF+1) <= xX(1) & "1";

  fpC(wE+wF) <= fpX(wE+wF) xor nY(2+wF+gK+5) xor nY(1+wF+gK+5);

  fpC(wE+wF-1 downto 0) <= eC & fC when nY(wF+gK+5) = nY(1+wF+gK+5) or eX(wE) = '1' else
                           eS & fS;

end architecture; -------------------------------------------------------------
